/*
 * Copyright © 2007 Global Locate.
 *
 * Written by Ken Yale.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include "nmea.h"

#include <gtk/gtk.h>
#include <glib/gprintf.h>

int gllin_fd();
extern void page_MAP_have_lla(double lon, double lat, int alt, int heading, int speed, int time);
extern FILE *dump_nmea_fp;
extern int debug_all;

void page_log_append(const char *nmea);

#define FIFO "/tmp/nmeaNP"
#define NV1  "./NVRAM1.DAT"
#define NV2  "./NVRAM2.DAT"
#define LLL  "log"

#undef SLOG

#ifdef SLOG
#define LOG  "/dev/log"
#endif

#define SIZE 128
#define GOT  3

FILE* fp1 = NULL;
int finish = 0;
char* gllin_stop(void);
gint agps_cb_nmea_tag = 0;
static void agps_nmea_process(GtkWidget* label, unsigned char* buf, int size);
unsigned long int runtime();
extern void page_SS_init();
extern void page_RFI_handle_GSV(nmea_gsv* pGSV);
extern void page_RFI_handle_GGA(nmea_gga* pGGA);
extern void pm_controls_nmea_epoch_done(void);
extern void page_stress_RID();
extern void page_stress_FIX();
extern void page_stress_EXIT();

//static char buffer_zone[1024];          // Somebody is overwriting a buffer.
static struct nmea_gga n_gga;
static struct nmea_rmc n_rmc;
static struct nmea_gsv n_gsv[GOT];
static struct nmea_gsa n_gsa;
static struct nmea_lor n_lor;

GtkLabel* agpsLabel = 0;

void NMEA_MSG(const char* msg)
{
    if (agpsLabel)
    {
        gtk_label_set_text(agpsLabel, msg);		
    }
}

static void NMEA_DEBUG(const char* msg)
{
#if 0
#ifdef DEBUG_LV1			
    printf("%s\n", msg);		
#endif				
#ifdef DEBUG_LV3
    NMEA_MSG(msg);
#endif	
#ifdef DEBUG_LV5
    print(msg);
#endif
#ifdef SLOG			
    if (fp1) fputs(msg, fp1);
#endif			
#endif
}

gboolean stop(GtkWidget *widget)
{
    finish = 1;
    return FALSE;
}

void
agps_fd(gpointer data, gint source, GdkInputCondition cond)
{
    //  For stress test only:
    //  Only needed if the g_timeout_cb() is not thread-safe:
    //
    //          page_main_poll_for_stop();
    //

    GtkWidget* label = GTK_WIDGET(data);
    if (source != gllin_fd())
    {
        NMEA_DEBUG("fd mismatch");
        return;
    }
    if (GDK_INPUT_READ != cond)
    {
        NMEA_DEBUG("gtk read callback condition mismatch");
        return;
    }
    unsigned char buf[SIZE+1];
    int n = read(source, buf, SIZE);
    if (n > 0)
	    agps_nmea_process(label, buf, n);
    else
	    printf("read failed\n");
}

void
agps_cb_start(GtkWidget* label)
{
    agps_cb_nmea_tag = gdk_input_add(gllin_fd(), GDK_INPUT_READ, agps_fd, label);
}

void
agps_cb_stop()
{
    gdk_input_remove(agps_cb_nmea_tag);
}

int fix_good  = 0;
int fix_fail  = 0;
int fix_total = 0;
static int ttff;

extern void page_MAP_init(void);

void
clearStats()
{
    fix_good  = 0;
    fix_fail  = 0;
    fix_total = 0;
    ttff = 0;
    page_SS_init();
    page_MAP_init();
}

static char line_buffer[1024] = { 0 };
static void agps_nmea_process_(GtkWidget* label, char* buf);

//-------------------------------------------------------------------------------------
//
//      agps_nmea_process()
//
//      Convert the chunks of NMEA data into separate sentences.
//      Then process.
//
//-------------------------------------------------------------------------------------

static unsigned int next_msg(GByteArray **pq)
{
	unsigned char *head, *tail, *end;
	GByteArray *q = *pq;

	if (q->len < 5)
		return FALSE;

	end = q->data + q->len;

	head = q->data;
	tail = end + 1;

	while (head < end)
	{
		if (*head == 0xb5)
		{
			if (head + 6 > end)
				break;

			if (head[1] == 0x62)
			{
				int len;

				len = head[4] | (head[5] << 8);
				/* limit payload size */
				if (len < 0x100)
				{
					tail = head + 6 + len + 2;

					//printf("ubx of len %d, got %d\n", len, q->len);

					break;
				}
			}
		}
		else if (*head == '$')
		{
			unsigned char *tmp = head + 1;

			while (g_ascii_isprint(*tmp) && tmp < end)
				tmp++;

			if (tmp + 2 > end)
				break;

			if ((tmp[0] == '\r' || tmp[0] == '\n') && tmp[1] == '\n')
			{
				tail = tmp + 2;

				break;
			}
			else
			{
				printf("nmea ends with 0x%02x 0x%02x\n", tmp[0], tmp[1]);
				head = tmp - 1;
			}
		}

		head++;
	}

	if (q->data < head)
	{

#if 0
		if (head - q->data == 20 && q->data[0] == 0x10 && q->data[1] == 0x00)
		{
			unsigned char nav_status[] = { 0xb5, 0x62, 0x01, 0x03 };

			q = g_byte_array_prepend(q, nav_status, sizeof(nav_status));

			head = q->data;
			tail = q->data + 24;
			end = q->data + q->len;
		}
		else
#endif
		{
			if (debug_all)
			{
				int i;

				printf("abandoning %d bytes of %d bytes\n---\n", head - q->data, q->len);

				for (i = 0; i < head - q->data; i++)
				{
					printf("0x%02x", q->data[i]);
					printf("%c", (i % 16 == 15) ? '\n' : ' ');
				}
				printf("\n---\n");
			}

			q = g_byte_array_remove_range(q, 0, head - q->data);
		}

		*pq = q;
	}

	return (tail <= end) ? (tail - head) : 0;
}

#define READ32(p)	(READ16(p) | READ16((p) + 2) << 16)
#define READ16(p)	((p)[0] | (p)[1] << 8)
#define READ8(p)	((p)[0])

static void ubx_process(unsigned char *buf, size_t size)
{
	unsigned char *p, cls, id, checksum[2];
	unsigned int i, payload;

	if (size < 6 || buf[0] != 0xb5 || buf[1] != 0x62)
	{
		printf("invalid ubx\n");

		return;
	}

	cls = buf[2];
	id = buf[3];

	payload = READ16(buf + 4);
	if (6 + payload + 2 != size)
	{
		printf("wrong ubx size\n");

		return;
	}

	checksum[0] = checksum[1] = 0;
	for (i = 2; i < size - 2; i++)
	{
		checksum[0] += buf[i];
		checksum[1] += checksum[0];
	}

	if (buf[size - 2] != checksum[0] || buf[size - 1] != checksum[1])
	{
		printf("wrong checksum (should be 0x%02x%02x, not 0x%02x%02x)\n",
				checksum[0], checksum[1],
				buf[size - 2], buf[size - 1]);

		return;
	}

	p = &buf[6];

	printf("received (0x%02x 0x%02x)\n", cls, id);

	switch (cls)
	{
	case 0x05:
		if (id == 0x01)
			printf("(0x%02x 0x%02x) acked\n", p[0], p[1]);
		break;
	case 0x01:
		switch (id)
		{
		case 0x03:
			printf("STATUS %u 0x%x 0x%x 0x%x %u %u\n",
					READ32(p) / 1000, READ8(p + 4), READ8(p + 5),
					READ8(p + 6), READ32(p + 8), READ32(p + 12));
			break;
		case 0x22:
			printf("TIME %u %d %d %u %u\n",
					READ32(p) / 1000, READ32(p + 4), READ32(p + 8),
					READ32(p + 12), READ32(p + 16));
			break;
		}
		break;
	}
}

static void
agps_nmea_process(GtkWidget* label, unsigned char* buf, int size)
{
    static GByteArray *q;
    int s;

    if (!q)
    {
	    q = g_byte_array_new();
	    if (!q)
		    return;
    }

    q = g_byte_array_append(q, buf, size);

    while ((s = next_msg(&q)))
    {
	    if (q->data[0] == '$')
	    {
		    memcpy(line_buffer, q->data, s - 2);
		    line_buffer[s - 2] = '\0';

		    agps_nmea_process_(label, line_buffer);
	    }
	    else
		    ubx_process(q->data, s);

	    q = g_byte_array_remove_range(q, 0, s);
    }
}

//------------------------------------------------------------------------------
//
//               nmea_to_ll(const char* s, float& lat)
//
//      Convert 12345.678 into
//      123 degrees, 45.678 minutes, or
//      123.761300 degrees.
//
//------------------------------------------------------------------------------

static float
nmea_to_ll(const char* s)
{
    bool  negate = false;
    if (*s == '-')
    {
        negate = true;
        ++s;
    }

    float fll;
    sscanf(s, "%f", &fll);              // 12345.678

    int   ill;
    sscanf(s, "%d", &ill);              // 12345
    float degrees = ill / 100;          // 123.0

    fll -= degrees * 100;               // 12345.678 - 12300.0 = 45.678
    degrees += fll/60.0;                // 123.0 + 45.678/60.0 = 123.7613
    if (negate)
    {
        degrees = -degrees;
    }
    return degrees;
}

extern char testStartTime[];

static char nmea_epoch_string[2048] = {0};

const char*
nmea_epoch()
{
    return nmea_epoch_string;
}

static void
nmea_epoch_end(struct nmea_gga* gga, struct nmea_lor* lor)
{
    char tmp[1024];
    char* p = tmp;
    memset(tmp, 0, sizeof(tmp));

    //p += sprintf(p, "Start time: %s\n", testStartTime);

    // TODO:  Report more info
    //        Be smarter about calling this procedure
    //        Save the data so we can press a "report" button to save to file.
    {
        unsigned long total_sec = runtime();
        p += sprintf(p, "Module Uptime:");

        if (total_sec > 59)
        {
            int hours =  total_sec / 3600;
            int min   = (total_sec % 3600) / 60;
            int sec   = (total_sec %   60);

            p += sprintf(p, " %02d:%02d:%02d", hours, min, sec);

        }
        p += sprintf(p, " (%d s)\n", (int) total_sec);
    }

    if (fix_total)
    {
        p += sprintf(p, "Fixes: %d good  %d fail   %d total\n",
                         fix_good, fix_fail, fix_total);
    }
    else
    {
        p += sprintf(p, "Starting\n");
    }

    if (ttff)
    {
	    p += sprintf(p, "TTFF: %ds\n", ttff);
    }
    else
    {
	    p += sprintf(p, "TTFF: -\n");
    }

    if (gga->time[0])
    {
        char fixTime[32];

        fixTime[0] = gga->time[0];
        fixTime[1] = gga->time[1];
        fixTime[2] = ':';
        fixTime[3] = gga->time[2];
        fixTime[4] = gga->time[3];
        fixTime[5] = ':';
        fixTime[6] = gga->time[4];
        fixTime[7] = gga->time[5];
        fixTime[8] = gga->time[6];
        fixTime[9] = gga->time[7];
        fixTime[10] = gga->time[8];
        fixTime[11] = 0;
        p += sprintf(p, "UTC: %s\n", fixTime);
    }
    else
    {
        p += sprintf(p, "Fix time: -\n");
    }

    if (strlen(gga->latitude) && strlen(gga->longitude))
    {
        float lat = nmea_to_ll(gga->latitude);
        float lon = nmea_to_ll(gga->longitude);

        int alt;
        sscanf(gga->alme, "%d", &alt);

        const char* alt_units = "m";

        p += sprintf(p,
                     "Latitude:  %10.6f %s\n"
                     "Longitude: %10.6f %s\n"
                     "Altitude:  %d %s\n"
                     "HDOP:      %s\n",
             lat, gga->la,
             lon, gga->lo,
             alt, alt_units,
             gga->hdop);

        if (strchr(gga->la, 'S'))
        {
            lat = -lat;
        }

        if (strchr(gga->lo, 'W'))
        {
            lon = -lon;
        }

        // TODO:  fill in alt, speed, and heading

        page_MAP_have_lla(lon, lat, 50, 0, 0, 123000);
    }
    else
    {
        p += sprintf(p,
                     "Latitude:  -\n"
                     "Longitude: -\n"
                     "Altitude:  -\n"
                     "HDOP:      -\n");
    }

    if (gga->nos[0])
    {
        int nsat;
        sscanf(gga->nos,  "%d", &nsat);
        p += sprintf(p, "Num sats: %d\n", nsat);
    }
    else
    {
        p += sprintf(p, "\n");
    }

    if (lor->rid[0])
    {
        p += sprintf(p, "Version: %s\n", lor->rid);
    }
    if (lor->nIgr)
    {
        p += sprintf(p, "IGR: %d) %s", lor->nIgr, lor->igr);
    }
    // g_printf("%s", tmp);
    NMEA_MSG(tmp);

    strcpy(nmea_epoch_string, tmp);

    pm_controls_nmea_epoch_done();
}

//-------------------------------------------------------------------------------------
//
//      agps_nmea_process_()
//
//      Process the NMEA sentences.
//
//-------------------------------------------------------------------------------------

static void
agps_nmea_process_(GtkWidget* label, char* buf)
{
    int j = 0;

    struct nmea_gga* gga = &n_gga;
    struct nmea_rmc* rmc = &n_rmc;
    struct nmea_gsv* gsv =  n_gsv;
    struct nmea_gsa* gsa = &n_gsa;
    struct nmea_lor* lor = &n_lor;

    NMEA_DEBUG(buf);

    page_log_append(buf);
    if (dump_nmea_fp)
	    g_fprintf(dump_nmea_fp, ">%s\n", buf);

    if (strstr(buf,GGA))
    {
        memset(gga, 0, sizeof(struct nmea_gga));
        NMEA_DEBUG("$GPGGA");
        GPGGA(buf, gga);						
        //page_RFI_handle_GGA(gga);

#ifdef DEBUG_LV2
        printf("Sentence: %s\n", gga->sentence);
        printf("FIX Taken: %s\n", gga->time);
        printf("Latitude: %s\n", gga->latitude);
        printf("%s\n", gga->la);
        printf("Longitude: %s\n", gga->longitude);
        printf("%s\n", gga->lo);
        printf("Fix quality: %s\n", gga->fix_quality);
        printf("Num of sat: %s\n", gga->nos);
        printf("Horizontal dilution: %s\n", gga->hdop);
        printf("Altitude: %s\n", gga->alme);
        printf("%s\n", gga->al);
        printf("Height of geoid: %s\n", gga->hog);
        printf("%s\n", gga->ho);
        printf("time in seconds: %s\n", gga->tis);
        printf("DGPS ID: %s\n", gga->DGPS);
        printf("checksum: %s\n", gga->checksum);
#endif
    }
    else if (strstr(buf,RMC))
    {
        memset(rmc, 0, sizeof(struct nmea_rmc));
        NMEA_DEBUG("$GPRMC");
        if (strstr(buf, ",A,"))
        {
            ++fix_good;
	    if (fix_good > 2 && !ttff)
		    ttff = runtime();
        }
        else
        {
            ++fix_fail;
        }
        ++fix_total;

        GPRMC(buf, rmc);	
        // g_printf("RMC: (%d+%d)/%d\n", fix_good, fix_fail, fix_total);
    }
    else if (strstr(buf,GSA))
    {
        memset(gsa, 0, sizeof(struct nmea_gsa));
        NMEA_DEBUG("$GPGSA");
        GPGSA(buf, gsa);
#ifdef DEBUG_LV2
        printf("Sentence: %s\n", gsa->sentence);
        printf("Auto/Manual: %s\n", gsa->AM);
        printf("mode: %s\n", gsa->mode);

        for (i = 0; i < 12; ++i) 
        {
            printf("PRN(%d): %s\n", i, gsa->id[i]);
        }

        printf("PDOP: %s\n", gsa->PDOP);
        printf("HDOP: %s\n", gsa->HDOP);
        printf("VDOP: %s\n", gsa->VDOP);
        printf("checksum: %s\n", gsa->checksum);
        printf("Num of Sate: %d\n",gsa->n);
#endif
        nmea_epoch_end(gga, lor);
    }
    else if (strstr(buf,GSV))
    {
        nmea_gsv* pGSV = gsv+j;
        memset(pGSV, 0, sizeof(struct nmea_gsv));
        NMEA_DEBUG("$GPGSV");
        GPGSV(buf, pGSV);	
#ifdef DEBUG_LV6
        printf("Sentence: %s\n",        pGSV->sentence);
        printf("Num of sentence: %s\n", pGSV->nose);
        printf("sentence #: %s\n",      pGSV->sn);
        printf("Num of sate: %s\n",     pGSV->nosa);

        for (i = 0; i < pGSV->n; ++i)
        {
            printf("(%d) PRN: %s\n",       i, pGSV->sate[i][0]); printf("(%d) Elevation: %s\n", i, pGSV->sate[i][1]);
            printf("(%d) Azimuth: %s\n",   i, pGSV->sate[i][2]);
            printf("(%d) SNR: %s\n",       i, pGSV->sate[i][3]);
        }

        printf("checksum: %s\n", (gsv+j)->checksum);
#endif

        //page_RFI_handle_GSV(pGSV);
        ++j;
    }
    else if (strstr(buf,"PGLOR,RID"))
    {
        memset(lor, 0, sizeof(struct nmea_lor));
        PGLOR_RID(buf, lor);
        page_stress_RID();
    }
    else if (strstr(buf,"PGLOR,IG"))
    {
        PGLOR_IGR(buf, lor);
    }
    else if (strstr(buf,"PGLOR,FIX"))
    {
        PGLOR_FIX(buf, lor);
        // nmea_epoch_end(gga, lor);
        page_stress_FIX();
    }
    else if (strstr(buf,"PGLOR,EXIT"))
    {
        PGLOR_EXIT(buf, lor);
        // nmea_epoch_end(gga, lor);
        page_stress_EXIT();
    }
}

void agps_stop(void)
{
    finish = TRUE;	
}


#if 0



#endif
